#include "../Main.h"

InputManager * g_InputManager = nullptr;

InputManager * InputManager::m_pInstance = nullptr;

InputManager::InputManager()
	: m_TypingPointer(),
	  m_MousePointer(),
	  m_TouchPointers(),
	  m_Listeners(),
	  m_PointerListeners(),
	  m_MouseCursorPosition(),
	  m_MouseCursorVisible(true),
	  m_MouseIgnorePositionOnce(true),
	  m_MouseIgnorePositionAlways(false),
	  m_WindowDimensions()
{
	assert(nullptr == m_pInstance);		// m_pInstance wasn't == nullptr, means we have more than 1 instance of InputManager, not right
	m_pInstance = this;

	SetGlfwCallbacks();
}

InputManager::~InputManager()
{
	// Clear all pointer mappings
	m_TypingPointer.GetWidgetMapping().RemoveAllMappings();
	m_MousePointer.GetWidgetMapping().RemoveAllMappings();
	for (auto & TouchPointer : m_TouchPointers)
	{
		TouchPointer.second->GetWidgetMapping().RemoveAllMappings();
	}

	RemoveGlfwCallbacks();

	SetMouseCursorVisibility(true);

	m_pInstance = nullptr;
}

void InputManager::RegisterListener(InputListener * Listener)
{
	m_Listeners.push_back(Listener);
}

void InputManager::UnregisterListener(InputListener * Listener)
{
	for (auto it0 = m_Listeners.begin(); it0 != m_Listeners.end(); ++it0)
	{
		if (*it0 == Listener)
		{
			m_Listeners.erase(it0);
			break;
		}
	}
}

void InputManager::RegisterPointerListener(PointerInputListener * PointerListener)
{
	m_PointerListeners.push_back(PointerListener);
}

void InputManager::UnregisterPointerListener(PointerInputListener * PointerListener)
{
	for (auto it0 = m_PointerListeners.begin(); it0 != m_PointerListeners.end(); ++it0)
	{
		if (*it0 == PointerListener)
		{
			m_PointerListeners.erase(it0);
			break;
		}
	}
}

const Vector2n InputManager::GetMouseCursorPosition() const
{
	return m_MouseCursorPosition;
}

void InputManager::SetMouseCursorVisibility(bool Visible)
{
	static int MouseCursorDesktopPositionX, MouseCursorDesktopPositionY;

	// Hide Mouse Cursor
	if (!Visible && m_MouseCursorVisible)
	{
		glfwPollEvents();
		m_MouseIgnorePositionAlways = true;
		glfwGetMousePos(&MouseCursorDesktopPositionX, &MouseCursorDesktopPositionY);
		glfwDisable(GLFW_MOUSE_CURSOR);
		m_MouseIgnorePositionOnce = true;
		m_MouseIgnorePositionAlways = false;

		m_MouseCursorVisible = false;
	}
	// Show Mouse Cursor
	else if (Visible && !m_MouseCursorVisible)
	{
		m_MouseIgnorePositionAlways = true;
		glfwEnable(GLFW_MOUSE_CURSOR);
		glfwSetMousePos(MouseCursorDesktopPositionX, MouseCursorDesktopPositionY);
		m_MouseIgnorePositionOnce = true;
		m_MouseIgnorePositionAlways = false;

		m_MouseCursorVisible = true;

		// Create a mouse position event
		//ProcessMousePos(m_MousePositionX, m_MousePositionY);
	}
}

bool InputManager::IsMouseCursorVisible()
{
	return m_MouseCursorVisible;
}

Vector2n InputManager::GetWindowDimensions()
{
	return m_WindowDimensions;
}

void InputManager::SetGlfwCallbacks()
{
	glfwSetWindowSizeCallback(&InputManager::ProcessWindowSize);

	glfwSetKeyCallback(&InputManager::ProcessKey);
	glfwSetCharCallback(&InputManager::ProcessChar);
	glfwSetMouseButtonCallback(&InputManager::ProcessMouseButton);
	glfwSetMousePosCallback(&InputManager::ProcessMousePos);
	glfwSetMouseWheelCallback(&InputManager::ProcessMouseWheel);
#ifdef _GLFW_DMITRI_WINDOWS_TOUCH_ENABLED
	glfwSetTouchButtonCallback(&InputManager::ProcessTouchButton);
	glfwSetTouchPosCallback(&InputManager::ProcessTouchPos);
#endif // _GLFW_DMITRI_WINDOWS_TOUCH_ENABLED
}

void InputManager::RemoveGlfwCallbacks()
{
	glfwSetWindowSizeCallback(nullptr);

	glfwSetKeyCallback(nullptr);
	glfwSetCharCallback(nullptr);
	glfwSetMouseButtonCallback(nullptr);
	glfwSetMousePosCallback(nullptr);
	glfwSetMouseWheelCallback(nullptr);
#ifdef _GLFW_DMITRI_WINDOWS_TOUCH_ENABLED
	glfwSetTouchButtonCallback(nullptr);
	glfwSetTouchPosCallback(nullptr);
#endif // _GLFW_DMITRI_WINDOWS_TOUCH_ENABLED
}

// Process a window resize event
void GLFWCALL InputManager::ProcessWindowSize(int WindowWidth, int WindowHeight)
{
	// Set the viewport size
	glViewport(0, 0, WindowWidth, WindowHeight);

	// Update the projection matrix
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0, WindowWidth, WindowHeight, 0);
	glMatrixMode(GL_MODELVIEW);

	m_pInstance->m_WindowDimensions = Vector2n(WindowWidth, WindowHeight);
	printf("Window Dimensions: %dx%d\n", m_pInstance->m_WindowDimensions.X(), m_pInstance->m_WindowDimensions.Y());

	// DEBUG: Hardcoded hack
	if (nullptr != pMainScreen)
		pMainScreen->UpdateWindowDimensions(m_pInstance->m_WindowDimensions);
}

void GLFWCALL InputManager::ProcessKey(int Key, int Action)
{
	m_pInstance->m_TypingPointer.ProcessButton(Key, (GLFW_PRESS == Action));

	for (std::vector<InputListener *>::reverse_iterator it0 = m_pInstance->m_Listeners.rbegin(); it0 != m_pInstance->m_Listeners.rend(); ++it0)
	{
		(*it0)->ProcessButton(InputId(0 /* Keyboard */, Key), (GLFW_PRESS == Action));
	}
}

void GLFWCALL InputManager::ProcessChar(int Character, int Action)
{
	if (GLFW_PRESS == Action)
	{
		m_pInstance->m_TypingPointer.ProcessCharacter(Character);
	}

	for (std::vector<InputListener *>::reverse_iterator it0 = m_pInstance->m_Listeners.rbegin(); it0 != m_pInstance->m_Listeners.rend(); ++it0)
	{
		if (GLFW_PRESS == Action)
		{
			(*it0)->ProcessCharacter(Character);
		}
	}
}

void GLFWCALL InputManager::ProcessMouseButton(int MouseButton, int Action)
{
	m_pInstance->m_MousePointer.ProcessButton(MouseButton, (GLFW_PRESS == Action));

	for (std::vector<PointerInputListener *>::reverse_iterator it0 = m_pInstance->m_PointerListeners.rbegin(); it0 != m_pInstance->m_PointerListeners.rend(); ++it0)
	{
		(*it0)->ProcessButton(&m_pInstance->m_MousePointer, Input::InputId(MouseButton), (GLFW_PRESS == Action));
	}
}

void GLFWCALL InputManager::ProcessMousePos(int MousePositionX, int MousePositionY)
{
#if 0
	static bool PreviousMousePositionSet = false;

	if (true == m_pInstance->IsMouseCursorVisible())
	{
		PreviousMousePositionSet = false;

		//m_pInstance->m_MousePositionX = MousePositionX;
		//m_pInstance->m_MousePositionY = MousePositionY;

		for (std::vector<InputListener *>::reverse_iterator it0 = m_pInstance->m_Listeners.rbegin(); it0 != m_pInstance->m_Listeners.rend(); ++it0)
		{
			//(*it0)->ProcessMousePosition(MousePositionX, MousePositionY);
			(*it0)->Process2Axes(InputId(1000 /* Mouse */, 0), MousePositionX, MousePositionY);
		}

		//printf("       pointer moved to = (%d, %d)\n", nMousePosX, nMousePosY);
	}
	else
	{
		static int		PreviousMousePositionX = MousePositionX;
		static int		PreviousMousePositionY = MousePositionY;

		if (false == PreviousMousePositionSet)
		{
			//printf("  initial mouse pos = (%d, %d)\n", nMousePosX, nMousePosY);
			PreviousMousePositionX = MousePositionX;
			PreviousMousePositionY = MousePositionY;
			PreviousMousePositionSet = true;
		}
		else
		{
			int MouseMovedX = MousePositionX - PreviousMousePositionX;
			int MouseMovedY = MousePositionY - PreviousMousePositionY;
			PreviousMousePositionX = MousePositionX;
			PreviousMousePositionY = MousePositionY;

			for (std::vector<InputListener *>::reverse_iterator it0 = m_pInstance->m_Listeners.rbegin(); it0 != m_pInstance->m_Listeners.rend(); ++it0)
			{
				(*it0)->ProcessSlider(InputId(1000 /* Mouse */, 0 /* Mouse X Axis */), static_cast<double>(MouseMovedX));
				(*it0)->ProcessSlider(InputId(1000 /* Mouse */, 1 /* Mouse Y Axis */), static_cast<double>(MouseMovedY));
			}
		}
	}
#else
	static int PreviousMousePositionX = MousePositionX;
	static int PreviousMousePositionY = MousePositionY;

	int MouseMovedX = MousePositionX - PreviousMousePositionX;
	int MouseMovedY = MousePositionY - PreviousMousePositionY;
	PreviousMousePositionX = MousePositionX;
	PreviousMousePositionY = MousePositionY;

	if (m_pInstance->m_MouseIgnorePositionAlways)
	{
		//printf("mm - %d,%d (Always)\n", MouseMovedX, MouseMovedY);
		return;
	}

	if (m_pInstance->m_MouseIgnorePositionOnce)
	{
		//printf("Mouse NOT moved by %d, %d to new pos %d, %d\n", MouseMovedX, MouseMovedY, MousePositionX, MousePositionY);
		m_pInstance->m_MouseIgnorePositionOnce = false;
		return;
	}

	if (true == m_pInstance->IsMouseCursorVisible())
	{
		m_pInstance->m_MouseCursorPosition.X() = MousePositionX;
		m_pInstance->m_MouseCursorPosition.Y() = MousePositionY;

		m_pInstance->m_MousePointer.ProcessPosition(m_pInstance->m_MouseCursorPosition);

		for (std::vector<PointerInputListener *>::reverse_iterator it0 = m_pInstance->m_PointerListeners.rbegin(); it0 != m_pInstance->m_PointerListeners.rend(); ++it0)
		{
			(*it0)->ProcessPosition(&m_pInstance->m_MousePointer, m_pInstance->m_MouseCursorPosition);
		}

		for (std::vector<InputListener *>::reverse_iterator it0 = m_pInstance->m_Listeners.rbegin(); it0 != m_pInstance->m_Listeners.rend(); ++it0)
		{
			//(*it0)->ProcessMousePosition(MousePositionX, MousePositionY);
			Input::AxisState AxisStates[2] = { Input::AxisState(MousePositionX, m_pInstance->m_WindowDimensions.X()), Input::AxisState(MousePositionY, m_pInstance->m_WindowDimensions.Y()) };
			(*it0)->Process2Axes(InputId(1000 /* Mouse */, 0), AxisStates);
		}
	}

	m_pInstance->m_MousePointer.ProcessSlider(0 /* Mouse X Axis */, static_cast<double>(MouseMovedX));
	m_pInstance->m_MousePointer.ProcessSlider(1 /* Mouse Y Axis */, static_cast<double>(MouseMovedY));

	for (std::vector<InputListener *>::reverse_iterator it0 = m_pInstance->m_Listeners.rbegin(); it0 != m_pInstance->m_Listeners.rend(); ++it0)
	{
		(*it0)->ProcessSlider(InputId(1000 /* Mouse */, 0 /* Mouse X Axis */), static_cast<double>(MouseMovedX));
		(*it0)->ProcessSlider(InputId(1000 /* Mouse */, 1 /* Mouse Y Axis */), static_cast<double>(MouseMovedY));
	}
#endif
}

void GLFWCALL InputManager::ProcessMouseWheel(int MouseWheelPosition)
{
	static int		PreviousMouseWheelPosition = MouseWheelPosition;

	int MouseWheelMoved = MouseWheelPosition - PreviousMouseWheelPosition;
	PreviousMouseWheelPosition = MouseWheelPosition;

	m_pInstance->m_MousePointer.ProcessSlider(2 /* Mouse Wheel */, MouseWheelMoved);

	for (std::vector<InputListener *>::reverse_iterator it0 = m_pInstance->m_Listeners.rbegin(); it0 != m_pInstance->m_Listeners.rend(); ++it0)
	{
		(*it0)->ProcessSlider(InputId(1000 /* Mouse */, 2 /* Mouse Wheel */), MouseWheelMoved);
		(*it0)->ProcessAxis(InputId(1000 /* Mouse */, 2 /* Mouse Wheel */), Input::AxisState(MouseWheelPosition, 0));
	}
}

#ifdef _GLFW_DMITRI_WINDOWS_TOUCH_ENABLED
void GLFWCALL InputManager::ProcessTouchButton(int TouchButton, int Action)
{
	m_pInstance->m_TouchPointers[TouchButton]->ProcessButton(0, (GLFW_PRESS == Action));

	for (std::vector<PointerInputListener *>::reverse_iterator it0 = m_pInstance->m_PointerListeners.rbegin(); it0 != m_pInstance->m_PointerListeners.rend(); ++it0)
	{
		(*it0)->ProcessButton(m_pInstance->m_TouchPointers[TouchButton].get(), Input::InputId(TouchButton), (GLFW_PRESS == Action));
	}

	if (GLFW_RELEASE == Action)
	{
		m_pInstance->m_TouchPointers.erase(m_pInstance->m_TouchPointers.find(TouchButton));
	}
printf("ProcessTouchButton(TouchButton %d, Action %d)\n", TouchButton, Action);
}

void GLFWCALL InputManager::ProcessTouchPos(int TouchButton, int TouchPositionX, int TouchPositionY, int Action)
{
	if (m_pInstance->m_TouchPointers.end() == m_pInstance->m_TouchPointers.find(TouchButton))
	{
		m_pInstance->m_TouchPointers.insert(std::pair<uint8, std::unique_ptr<TouchPointer>>(TouchButton, std::unique_ptr<TouchPointer>(new TouchPointer())));
	}

	auto & PreviousTouchPosition = m_pInstance->m_TouchPointers[TouchButton]->PreviousTouchPosition();

	int TouchMovedX = TouchPositionX - PreviousTouchPosition.X();
	int TouchMovedY = TouchPositionY - PreviousTouchPosition.Y();
	PreviousTouchPosition.X() = TouchPositionX;
	PreviousTouchPosition.Y() = TouchPositionY;

	Vector2n TouchPosition(TouchPositionX, TouchPositionY);

	m_pInstance->m_TouchPointers[TouchButton]->ProcessPosition(TouchPosition);

	for (std::vector<PointerInputListener *>::reverse_iterator it0 = m_pInstance->m_PointerListeners.rbegin(); it0 != m_pInstance->m_PointerListeners.rend(); ++it0)
	{
		(*it0)->ProcessPosition(m_pInstance->m_TouchPointers[TouchButton].get(), TouchPosition);
	}

	if (1 == Action)
	{
		m_pInstance->m_TouchPointers[TouchButton]->ProcessSlider(0 /* Touch X Axis */, static_cast<double>(TouchMovedX));
		m_pInstance->m_TouchPointers[TouchButton]->ProcessSlider(1 /* Touch Y Axis */, static_cast<double>(TouchMovedY));
	}

printf("ProcessTouchPos(TouchButton %d, X %d, Y %d, Action %d)\n", TouchButton, TouchPositionX, TouchPositionY, Action);
}
#endif // _GLFW_DMITRI_WINDOWS_TOUCH_ENABLED

void InputManager::ProcessTimePassed(const double TimePassed)
{
	m_TypingPointer.ProcessTimePassed(TimePassed);
	m_MousePointer.ProcessTimePassed(TimePassed);

	for (std::vector<PointerInputListener *>::reverse_iterator it0 = m_pInstance->m_PointerListeners.rbegin(); it0 != m_pInstance->m_PointerListeners.rend(); ++it0)
	{
		(*it0)->ProcessTimePassed(TimePassed);
	}
}
